import { gql } from "@amplicode/gql";
import {
  CheckAuthenticatedQuery,
  CheckAuthenticatedQueryVariables,
  UserInfoQuery,
  UserInfoQueryVariables,
  UserPermissionsQuery,
  UserPermissionsQueryVariables,
} from "@amplicode/gql/graphql";
import { ApolloQueryResult, TypedDocumentNode } from "@apollo/client";
import { GraphQLError } from "graphql";
import { AuthProvider } from "react-admin";
import { LOGIN_URI, LOGOUT_URI } from "../config";
import { apolloClient } from "../core/apollo/client";

export const PERMISSIONS_KEY: "amplicode-permissions" = "amplicode-permissions";

const USER_INFO: TypedDocumentNode<UserInfoQuery, UserInfoQueryVariables> = gql(`
  query userInfo {
   userInfo {
     id
     fullName
     avatar
   }
  }
`);

const USER_AUTHENTICATED: TypedDocumentNode<
  CheckAuthenticatedQuery,
  CheckAuthenticatedQueryVariables
> = gql(`
 query checkAuthenticated {
   checkAuthenticated
 }
`);

const USER_PERMISSIONS: TypedDocumentNode<UserPermissionsQuery, UserPermissionsQueryVariables> =
  gql(`
  query userPermissions {
   userPermissions
  }
`);

interface UserInfo {
  id: string;
  fullName?: string | null;
  avatar?: string | null;
}

// TODO importing from 'ra-core/src/types' break `npm run build` command
// import {UserIdentity} from "ra-core/src/types";
interface UserIdentity {
  id: string | number;
  fullName?: string;
  avatar?: string;
  [key: string]: any;
}

export const authProvider: AuthProvider = {
  login: () => Promise.reject("authProvider:login not used in this type of auth"),
  logout: async () => {
    localStorage.removeItem(PERMISSIONS_KEY);
    const url: string = composeUrl(LOGOUT_URI);
    console.log(`authProvider:logout redirect to ${url}`);
    window.location.href = url;
  },
  checkError: async (error) => {
    // TODO code below assumes that GraphQL server returns
    // {"errors":[{"extensions":{"classification":"UNAUTHORIZED"}}], ...}
    // for not authenticated user
    // and
    // {"errors":[{"extensions":{"classification":"FORBIDDEN"}}], ...}
    // if user has not enough permissions for query.
    // If the server handles errors differently, or has a different response structure,
    // code below should be modified.

    const { graphQLErrors, networkError } = error;
    // redirect to login page if graphql returns UNAUTHORIZED error
    if (
      Array.isArray(graphQLErrors) &&
      graphQLErrors.some((err: GraphQLError) => err.extensions?.classification === "UNAUTHORIZED")
    ) {
      const url = composeUrl(LOGIN_URI);
      window.location.href = url;
      console.log(`authProvider:checkError redirect to '${url}'`);
    }

    // redirect to login page for network error with 401 status
    if (networkError?.statusCode === 401) {
      const url = composeUrl(LOGIN_URI);
      window.location.href = url;
      console.log(`authProvider:checkError redirect to '${url}'`);
    }

    console.log("authProvider:checkError - not handle this type of error");
    return Promise.resolve();
  },
  checkAuth: async (): Promise<void> => {
    await apolloClient.query({ query: USER_AUTHENTICATED }).catch((e) => {
      // redirect to login page if graphql returns UNAUTHORIZED error
      if (isUnauthorized(e.graphQLErrors)) {
        const url: string = composeUrl(LOGIN_URI);
        console.log(`authProvider:checkAuth error 'UNAUTHORIZED', redirect to '${url}'`);
        return (window.location.href = url);
      }
      if (e.name === "ApolloError" && e.message === "Failed to fetch") {
        // do not perform logout in case of request cancel
        // in future we can add special error in apollo client for cancel requests
        // and handle it here more precise
        // fix for https://github.com/Amplicode/amplicode-frontend/issues/1149
        console.log("checkAuth: looks like request for auth check canceled, do not perform logout");
      } else {
        console.log(`authProvider:checkAuth throws `, JSON.stringify(e));
        throw e;
      }
    });
  },

  getPermissions: async (): Promise<string | null> => {
    let permissions: string | null = localStorage.getItem(PERMISSIONS_KEY);
    if (permissions != null) {
      return JSON.parse(permissions);
    }

    const authorities: ApolloQueryResult<UserPermissionsQuery> = await apolloClient.query({
      query: USER_PERMISSIONS,
    });

    permissions = authorities?.data?.userPermissions as string | null;
    if (permissions != null) {
      localStorage.setItem(PERMISSIONS_KEY, JSON.stringify(permissions));
    }
    return permissions;
  },

  getIdentity: async (): Promise<UserIdentity> => {
    const response: ApolloQueryResult<UserInfoQuery> = await apolloClient.query({
      query: USER_INFO,
    });
    const userInfo: UserInfo | null | undefined | void = response?.data?.userInfo;
    return {
      id: userInfo?.id ?? "",
      fullName: userInfo?.fullName ?? "",
      avatar: userInfo?.avatar ?? "",
    };
  },
};

function composeUrl(url: string): string {
  return `${url}?redirect_uri=${window.location.href}`;
}

function isUnauthorized(graphQLErrors: GraphQLError[]): boolean {
  return (
    Array.isArray(graphQLErrors) &&
    graphQLErrors.some((err: GraphQLError) => err.extensions?.classification === "UNAUTHORIZED")
  );
}
